from requests.auth import AuthBase
import logging
import os
from datetime import datetime, timedelta

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - [%(filename)s:%(lineno)d] - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[
        logging.FileHandler("blackduck.log"),
        logging.StreamHandler(),
    ],
)

logger = logging.getLogger(__name__)


class BearerAuth(AuthBase):
    """Authenticate with Blackduck hub using access token"""

    def __init__(self, session):
        """
        Args:
            session (requests.session): requests session to authenticate
        """

        self.blackduck_url = os.environ.get("BCOM_BDUCK_URL")
        self.api_token = os.environ.get("BCOM_BDUCK_API_TOKEN")
        self.session = session
        self.bearer_token = None
        self.csrf_token = None
        self.valid_until = datetime.now()

    def __call__(self, request):
        if not self.bearer_token or datetime.now() > self.valid_until - timedelta(
            minutes=5
        ):
            self.authenticate()

        request.headers.update(
            {
                "authorization": f"bearer {self.bearer_token}",
                "X-CSRF-TOKEN": self.csrf_token,
            }
        )

        return request

    def authenticate(self):
        logger.info(f"Authenticating to BlackDuck API ..")
        response = self.session.post(
            url=f"{self.blackduck_url}/api/tokens/authenticate",
            headers={"Authorization": f"token {self.api_token}"},
        )

        response.raise_for_status()

        content = response.json()
        self.bearer_token = content["bearerToken"]
        self.csrf_token = response.headers["X-CSRF-TOKEN"]
        self.valid_until = datetime.now() + timedelta(
            milliseconds=int(content["expiresInMilliseconds"])
        )

        logger.info(f"Auth granted until {self.valid_until.astimezone()}")

        return
