import logging
import os
import requests
import time
from authentication import BearerAuth

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - [%(filename)s:%(lineno)d] - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[
        logging.FileHandler("blackduck.log"),
        logging.StreamHandler(),
    ],
)

logger = logging.getLogger(__name__)


class BlackDuckCore:
    def __init__(self):
        self.blackduck_url = os.environ.get("BCOM_BDUCK_URL")
        self.session = requests.Session()
        self.auth = BearerAuth(self.session)

    def _make_request(self, method, url, **kwargs):
        """Helper method to make HTTP requests with retries."""
        retries = 3
        sleep_duration = 30

        for attempt in range(retries):
            try:
                response = self.session.request(method, url, **kwargs)

                response.raise_for_status()

                return response
            except (
                requests.exceptions.HTTPError,
                requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
            ) as e:
                if isinstance(e, requests.exceptions.HTTPError):
                    logger.error(f"HTTP error occurred: {e.response.status_code} - {e.response.text}")
                else:
                    logger.warning(f"Connection error occurred: {e}.")

                if attempt < retries - 1:
                    logger.info(f"Retrying request in {sleep_duration} seconds ..")
                    time.sleep(sleep_duration)
                else:
                    logger.error("Max retries reached. Aborting.")
                    raise
            except requests.exceptions.RequestException as e:
                logger.error(f"Request failed: {e}")
                raise

        logger.error("Max retries reached. Aborting.")
        raise requests.exceptions.RequestException("Max retries reached for the request.")

    def get_project_id(self, project_name):
        url = f"{self.blackduck_url}/api/projects?q=name:{project_name}"

        logger.info(f"Retrieving project ID for {project_name} ..")
        response = self._make_request("GET", url, auth=self.auth)
        projects = response.json().get("items", [])

        if projects:
            project_id = projects[0]["_meta"]["href"].split("/")[-1]
            logger.info(f"Project ID successfully retrieved: {project_id}")

            return project_id
        else:
            logger.error(f"Project {project_name} not found!")

            return None

    def get_projects(self, limit=1, parameters={}):
        if limit:
            parameters.update({"limit": limit})

        url = f"{self.blackduck_url}/api/projects"
        headers = {"Accept": "application/vnd.blackducksoftware.project-detail-4+json"}

        logger.info(f"Retrieving projects using url {url} with parameters {parameters}")
        response = self._make_request("GET", url, headers=headers, auth=self.auth)

        return response.json()

    def get_project_by_name(self, project_name):
        project_list = self.get_projects(parameters={"q": f"name:{project_name}"})

        for project in project_list.get("items", []):
            if project["name"] == project_name:
                return project

        logger.warning(f"Project {project_name} not found in the project list.")

        return None

    def project_exists(self, project_name):
        url = f"{self.blackduck_url}/api/projects?q=name:{project_name}"

        logger.info(f"Checking if project {project_name} exists in BlackDuck ..")
        response = self._make_request("GET", url, auth=self.auth)

        if response.json().get("totalCount", 0) > 0:
            logger.info(f"{project_name} exists in BlackDuck")
            return True

        return False

    def get_project_versions(self, project_id, project_name):
        url = f"{self.blackduck_url}/api/projects/{project_id}/versions"

        logger.info(f"Retrieving project versions for {project_name} ..")
        response = self._make_request("GET", url, auth=self.auth)
        versions = response.json().get("items", [])

        if versions:
            logger.info(f"Project versions successfuly retrieved for {project_name}")
            return versions
        else:
            return None

    def get_project_distribution(self, version_url, project_version_name):
        logger.info(f"Retrieving project distribution for {project_version_name} ..")
        response = self._make_request("GET", version_url, auth=self.auth)

        if response.status_code == 200:
            project_distribution = response.json().get("distribution", "EXTERNAL")
            logger.info(f"Project distribution successfully retrieved: {project_distribution}")

            return project_distribution
        else:
            logger.error(f"Failed to retrieve project distribution. Status code: {response.status_code}, Response: {response.text}")
            raise Exception(f"Failed to retrieve project distribution for '{project_version_name}': {response.text}")

    def update_version_phase(self, project_id, version_url, project_version_name, phase):
        distribution = self.get_project_distribution(version_url, project_version_name)

        data = {
            "phase": phase,
            "versionName": project_version_name,
            "distribution": distribution,
        }

        logger.info(f"Updating project version phase for {project_version_name} to {phase} ..")
        response = self._make_request("PUT", version_url, json=data, auth=self.auth)

        if response.status_code == 200:
            versions = self.get_project_versions(project_id, project_version_name)

            for version in versions:
                if version["versionName"] == project_version_name:
                    if version["phase"] == phase:
                        logger.info(f"Project version phase changes have been verified.")
                        return
                    else:
                        logger.error(f"Verification failed for project '{project_version_name}'! Expected phase '{phase}', but got '{version['phase']}'.")
                        raise Exception(f"Phase mismatch for '{project_version_name}': Expected '{phase}', found '{version['phase']}'.")

        logger.error(f"Failed to update project version. Status code: {response.status_code}, Response: {response.text}")
        raise Exception(f"Failed to update project version: {response.text}")
