#! /usr/bin/env python3

"""PSEC BlackDuck Scanner.

This script uses BlackDuck to perform licensing,
vulnerabilities, and binary analysis against files and docker images.
"""

import argparse
import logging
import os
import re
import requests
import shutil
import subprocess
import sys
import yaml
from requests.auth import HTTPBasicAuth
from urllib.parse import urlparse
from core import BlackDuckCore

REP_PATH_CLI = os.environ.get("REP_PATH_CLI", ".")
BCOM_BDUCK_URL = os.environ.get("BCOM_BDUCK_URL")
BCOM_BDUCK_API_TOKEN = os.environ.get("BCOM_BDUCK_API_TOKEN")
CONTAINER_PROJECTS = [
    "ESG-CBC-Platform",
    "CBC-DevRel-Extensions",
    "ESG-Product-Security-Test",
]

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - [%(filename)s:%(lineno)d] - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[
        logging.FileHandler("blackduck.log"),
        logging.StreamHandler(),
    ],
)

logger = logging.getLogger(__name__)


def validate_variables():
    """Validates the presence of required environment variables.

    Raises:
        ValueError: if any required environment variable is not set.
    """

    required_vars = ["BCOM_BDUCK_URL", "BCOM_BDUCK_API_TOKEN"]

    logger.info("Validating required variables ..")

    for var in required_vars:
        if not os.environ.get(var):
            logger.error(f"Required environment variable {var} is not set.")
            sys.exit(1)

    logger.info("All required environment variables are set.")


def download_file(url: str):
    """Downloads target file to be scanned.

    Args:
        url (str): URL of the file to be downloaded.

    Raises:
        RequestException: if the request fails.

    Returns:
        file_path (str): path to the file to be scanned.
    """

    session = requests.Session()
    auth = None

    try:
        if "usw1.packages.broadcom.com" in url:
            auth = HTTPBasicAuth(
                os.getenv("ARTIFACTORY_USER"), os.getenv("ARTIFACTORY_PASSWORD")
            )

        file_path = os.path.join(os.getcwd(), os.path.basename(urlparse(url).path))

        with session.get(url, auth=auth, stream=True) as s_response:
            with open(file_path, "wb") as file:
                for chunk in s_response.iter_content(chunk_size=8192):
                    file.write(chunk)

        logger.info(f"File downloaded and saved at: {file_path}")
        return file_path
    except requests.exceptions.RequestException as e:
        logger.error(f"Failed to download the target file: {e}")
        sys.exit(1)


def read_yaml(project_groups, project_name: str):
    """Traverses the dictionary and returns the project name or None.

    Args:
        project_groups (dict): dict containing all project groups.
        project_name (str): the name of the project to search for.

    Returns:
        project_group (str): the project group name.
    """

    if isinstance(project_groups, dict):
        for key, value in project_groups.items():
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        result = read_yaml(item, project_name)

                        if result:
                            return result
                    elif project_name.lower() in item.lower():
                        return key
            elif isinstance(value, dict):
                result = read_yaml(value, project_name)

                if result:
                    return result
    return None


def get_project_group(project_name: str):
    """Determines the project group name.

    Searches for project_name in the mapping file.
    If no matches are found, project_name is set to a default value.

    Args:
        project_name (str): the name of the project to search for.

    Raises:
        Exception: if something fails.

    Returns:
        project_group (str): the project group name.
    """

    mapping_file = os.path.join(os.getcwd(), "blackduck/projects_mapping.yml")

    try:
        with open(mapping_file, "r") as f:
            project_groups = yaml.safe_load(f)

        project_group = read_yaml(project_groups, project_name)

        if project_group is None:
            project_group = "CB Cloud (CBC)"
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        sys.exit(1)

    return project_group


def run_synopsys_detect(
    project_name: str,
    project_version: str,
    project_group: str,
    DETECT_TOOLS: str,
    EXCLUDED_TOOLS: str,
    LOGGING_LEVEL: str,
    target: str,
):
    """Runs BlackDuck Detect via shell command.

    Invokes BlackDuck Detect and scans a given target
    using several different detectors.

    Args:
        project_name (str): name of the project in BlackDuck HUB.
        project_version (str): version of the project in BlackDuck HUB.
        project_group (str): name of the project group in BlackDuck HUB.
        DETECT_TOOLS (str): Detect tools to include in the scan.
        EXCLUDED_TOOLS (str): Detect tools to exclude from the scan.
        target (str): file, directory or container image to be scanned.

    Raises:
        CalledProcessError: If the subprocess command fails.

    Returns:
        file_path (str): Path to the file to be scanned.
    """

    BDUCK_LOGGING_LEVEL = LOGGING_LEVEL or "INFO"
    BDUCK_DETECT_TOOLS = DETECT_TOOLS or "BINARY_SCAN"
    BDUCK_CLONE_CATEGORIES = "COMPONENT_DATA,CUSTOM_FIELD_DATA,DEEP_LICENSE,LICENSE_TERM_FULFILLMENT,VULN_DATA"
    BDUCK_EXCLUDED_TOOLS = (
        EXCLUDED_TOOLS
        or "DETECTOR,SIGNATURE_SCAN,IMPACT_ANALYSIS,DOCKER,BAZEL,IAC_SCAN,CONTAINER_SCAN,THREAT_INTEL"
    )
    blackduck_core = BlackDuckCore()
    project_name = project_name or "unknown"
    project_version = project_version or "unknown"
    project_group = project_group or get_project_group(project_name)
    target = target

    image_name_pattern = re.compile(
        r"^([a-z0-9.-]+(:[a-zA-Z0-9.-]+)?(/[a-zA-Z0-9.-]+)+(:[a-zA-Z0-9.-]+)?)$"
    )

    if not blackduck_core.project_exists(project_name):
        logger.error(f"The project '{project_name}' does not exist in BlackDuck. Exiting ..")
        sys.exit(1)

    cmd_args = {
        "--blackduck.url": BCOM_BDUCK_URL,
        "--blackduck.api.token": BCOM_BDUCK_API_TOKEN,
        "--blackduck.trust.cert": "true",
        "--detect.project.group.name": project_group,
        "--detect.project.name": project_name,
        "--detect.project.version.name": project_version,
        "--detect.tools": BDUCK_DETECT_TOOLS,
        "--detect.tools.excluded": BDUCK_EXCLUDED_TOOLS,
        "--logging.level.com.synopsys.integration": BDUCK_LOGGING_LEVEL,
        "--detect.risk.report.pdf": "false",
        "--detect.wait.for.results": "false",
        "--detect.timeout": "1200",
        "--detect.cleanup": "true",
        "--detect.clone.project.version.latest": "true",
        "--detect.project.clone.categories": BDUCK_CLONE_CATEGORIES,
        "--detect.project.version.update": "true",
    }

    if os.path.isdir(target):
        logger.error(f"Provide the full path to the binary file. A path to a directory was provided instead.")
        sys.exit(1)
    elif os.path.isfile(target):
        destination_path = os.path.join(REP_PATH_CLI, os.path.basename(target))

        if not os.path.exists(destination_path):
            try:
                target = shutil.copy2(target, f"{REP_PATH_CLI}/")
                cmd_args["--detect.binary.scan.file.path"] = target
            except Exception as e:
                logger.error(f"Failed to copy the target file: {e}")
                sys.exit(1)
        else:
            logger.info("Target file already exists in the destination. Proceeding ..")
            cmd_args["--detect.binary.scan.file.path"] = destination_path
    elif target.startswith(("http://", "https://")):
        target = download_file(target)

        try:
            target = shutil.copy2(target, f"{REP_PATH_CLI}/")
            cmd_args["--detect.binary.scan.file.path"] = target
        except Exception as e:
            logger.error(f"Failed to copy the target file: {e}")
            sys.exit(1)
    elif image_name_pattern.match(target):
        if "unknown" in project_version.lower() or project_version == "" or project_version is None:
            project_version = target.rsplit(":", 1)[0]
            cmd_args["--detect.project.version.name"] = project_version

        file_path = os.path.join(os.getcwd(), os.path.basename(project_version))
        cmd_args["--detect.project.version.distribution"] = "SAAS"

        if "binary_scan" in BDUCK_DETECT_TOOLS.lower():
            cmd_args["--detect.binary.scan.file.path"] = f"{file_path}.tar"
        if "container_scan" in BDUCK_DETECT_TOOLS.lower():
            cmd_args["--detect.container.scan.file.path"] = f"{file_path}.tar"
        if "docker" in BDUCK_DETECT_TOOLS.lower():
            cmd_args["--detect.docker.tar"] = f"{file_path}.tar"

        try:
            logger.info(f"Pulling {target}")
            subprocess.run(["docker", "pull", f"{target}"], check=True)

            logger.info(f"Saving {target} as a tar file")
            subprocess.run(["docker", "save", "--output", f"{file_path}.tar", f"{target}"], check=True)
        except Exception as e:
            logger.error(f"Command execution failed. Error: {e}")
            sys.exit(1)
    else:
        logger.error("Provide a file URL, local file path, or container image name.")
        sys.exit(1)

    cmd = ["java", "-jar", "/root/synopsys-detect.jar"]

    for arg, value in cmd_args.items():
        cmd.append(f"{arg}={value}")

    try:
        subprocess.run(cmd, check=True)
    except Exception as e:
        logger.error(f"Command execution failed. Error: {e}. Command: {' '.join(cmd)}")
        sys.exit(1)

    try:
        project_id = blackduck_core.get_project_id(project_name)
        versions = blackduck_core.get_project_versions(project_id, project_name)

        if not versions:
            logger.info("The project does not contain any versions. Skipping version deprecation and update.")
            sys.exit(0)

        versions.sort(key=lambda v: v["createdAt"])

        if project_name not in CONTAINER_PROJECTS:
            released_versions = [v for v in versions if v["phase"].lower() == "released"]

            if released_versions:
                released_versions.sort(key=lambda v: v["createdAt"])

                for rv in released_versions:
                    released_version_url = rv["_meta"]["href"]
                    released_version_name = rv["versionName"]

                    blackduck_core.update_version_phase(
                        project_id,
                        released_version_url,
                        released_version_name,
                        "DEPRECATED",
                    )
            else:
                logger.info("No versions in RELEASED phase found! Skipping version deprecation.")

        latest_version_url = versions[-1]["_meta"]["href"]
        latest_version_name = versions[-1]["versionName"]

        if versions[-1]["phase"].lower() != "released":
            blackduck_core.update_version_phase(project_id, latest_version_url, latest_version_name, "RELEASED")
        else:
            logger.info(f"The latest version '{latest_version_name}' is already marked as RELEASED.")
    except Exception as e:
        logger.error(f"Failed to retrieve project versions or update phases: {e}")
        sys.exit(1)


def main():
    validate_variables()

    parser = argparse.ArgumentParser(description="PSEC BlackDuck Scanner")

    parser.add_argument(
        "-pn",
        "--project_name",
        help="Project name in BlackDuck",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-pv",
        "--project_version",
        help="Project version in BlackDuck",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-pg",
        "--project_group",
        help="Project parent group name in BlackDuck",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-dt",
        "--detect_tools",
        help="Detect tools to include in the scan",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-et",
        "--excluded_tools",
        help="Detect tools to exclude from the scan",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-t",
        "--target",
        help="File URL, local file path, or full container image URL",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-ll",
        "--logging_level",
        help="BlackDuck logging level",
        required=False,
        type=str,
    )

    args = parser.parse_args()
    project_name = args.project_name
    project_version = args.project_version
    project_group = args.project_group
    DETECT_TOOLS = args.detect_tools
    EXCLUDED_TOOLS = args.excluded_tools
    LOGGING_LEVEL = args.logging_level
    target = args.target

    logger.info(f"Initiating BlackDuck scan activities ..")

    for t in target.split():
        run_synopsys_detect(
            project_name,
            project_version,
            project_group,
            DETECT_TOOLS,
            EXCLUDED_TOOLS,
            LOGGING_LEVEL,
            t,
        )

    logger.info("BlackDuck scan activities are now complete.")


if __name__ == "__main__":
    main()
